require("./moment-timezone-with-data-1970-2030.js");
const moment = require("./moment.js");

/*
   General utility class for manipulating timestamps based on their actual
   time-zone aware values for dates.

   The functions getTimestamp and getDateInfo use 1-based months, so
   January = 1, February = 2, etc., unlike javascript's zero-based months.

   This class uses moment.js and moment timezone but doesn't expose moments
   outside of the private functions. After we upgrade v8 we can replace moment
   calcs with something better (luxon, perhaps)
*/

class DateCalc
{
    constructor(timeZone = 'UTC', baseTimestamp = null, durationUnit = null)
    {
        this._timeZone = timeZone;
        this._setDurationUnit(durationUnit);
        this._baseTimestamp = baseTimestamp;
    }
    getDurationUnit() { return this._durationUnit; }
    _setDurationUnit(durationUnit)
    {
        this._durationUnit = durationUnit || 'months';
        switch (this._durationUnit)
        {
            case 'days360':
                this.getDuration = this.dayCount360;
                break;
            case 'months':
                this.getDuration = this.monthCount;
                break;
            case 'socotraMonths':
                this.getDuration = this.socotraMonthCount;
                break;
            case 'days':
                this.getDuration = this.dayCount;
                break;
            case 'wholeDays':
                this.getDuration = this.dayCountWhole;
                break;
            case 'ms':
                this.getDuration = (start, end) => end - start;
                break;
            default:
                throw `Duration unit ${durationUnit} not supported!`;
        }
    }
    /* one based months */
    getTimestamp(year = 1970, month = 1, day = 1, hour = 0, minute = 0, second = 0, millisecond = 0)
    {
        if (month < 1 || month > 12)
            throw `getTimestamp: month ${month} out of range. Use 1-based month values.`;

        return moment.tz( { year, month: month - 1, day, hour, minute, second, millisecond}, this._timeZone).valueOf();
    }
     /* one based months */
    getDateInfo(timestamp)
    {
        const m = this._toMoment(timestamp);
        return {
            year: m.year(),
            month: m.month() + 1,
            day: m.date(),
            hour: m.hour(),
            minute: m.minute(),
            second: m.second(),
            millisecond: m.millisecond()
        };
    }
    isDateInfoSame(info1, info2)
    {
        return info1.year === info2.year &&
               info1.month === info2.month &&
               info1.day === info2.day &&
               info1.hour === info2.hour &&
               info1.minute === info2.minute &&
               info1.second === info2.second &&
               info1.millisecond === info2.millisecond;
    }
    formatTimestamp(timestamp)
    {
        const di = this.getDateInfo(timestamp);
        return `${di.year}-${this._leadZero(di.month)}-${this._leadZero(di.day)} ${this._leadZero(di.hour)}:${this._leadZero(di.minute)}:${this._leadZero(di.second)}`;
    }
    formatTimestampDate(timestamp)
    {
        const di = this.getDateInfo(timestamp);
        return `${di.year}-${this._leadZero(di.month)}-${this._leadZero(di.day)}`;
    }
    _leadZero(num)
    {
        return ("00" + num.toString()).slice(-2);
    }
    // startTimestamp: inclusive lower bound of sequence
    // endTimestamp: exclusive upper bound of sequence
    // Options:
    //  increment: 'month', 'week', 'quarter', etc. Default is 'month'
    //  anchorTimestamp: ensure the sequence includes this timestamp. Default is startTimestamp. This will
    //                   be filtered out if it is outside of [start, end). Default is startTimestamp.
    //  maxCount: the maximum number of sequence items to return. Default is unlimited.
    //  returnMetadata: include information about the bounds of the sequence. Default is true
    //  returnIntervals: instead of a list of timestamps, return a list of { startTimestamp, endTimestamp }.
    //                   Default is true
    // increment can be an array of integers representing day offsets. If this is used the anchorTimestamp
    //     must be the same as the startTimestamp.
    getDateSequence(startTimestamp, endTimestamp, options = null)
    {
        // Can just pass the increment instead of an options object
        let increment = (typeof options === 'string') ? options : (options.increment || 'month');
        const anchorTimestamp = options.anchorTimestamp || startTimestamp;
        const maxCount = (increment === 'eon') ? 1 : (options.maxCount || 100000000);
        const returnMetadata = options.returnMetadata !== false;
        const returnIntervals = options.returnIntervals !== false;
   
        const cursor = this._toMoment(anchorTimestamp);
        
        const baseDay = cursor.date();
        let sequence = [];
        
        let incrementArray;

        if (startTimestamp >= endTimestamp)
            throw "startTimestamp must be before endTimestamp";

        if (Array.isArray(increment))
        {
            if (startTimestamp !== anchorTimestamp)
                throw "Must have startTimestamp equal anchorTimestamp when using day arrays.";
            incrementArray = increment;
            increment = 'arrayDays';
        }
        else if (increment === 'eon')
        {
            if (startTimestamp !== anchorTimestamp)
                throw "Must have startTimestamp equal anchorTimestamp when using eon interval";
        }
        else
        {
            while (startTimestamp < cursor.valueOf())
            {
                switch (increment)
                {
                    case '30day':
                        cursor.add(-30, 'days');
                        break;
                    case 'month':
                        this._incrementMomentByMonth(cursor, baseDay, -1);
                        break;
                    case 'quarter':
                        this._incrementMomentByMonth(cursor, baseDay, -3);
                        break;
                    case 'halfYear':
                        this._incrementMomentByMonth(cursor, baseDay, -6);
                        break;
                    case 'year':
                        cursor.add(-1, 'years');
                        break;
                    case 'week':
                        cursor.add(-7, 'days');
                        break;
                    case '2week':
                        cursor.add(-14, 'days');
                        break;
                    default:
                        throw `Span increment '${increment}' not supported!`;
                }
            }
        }
        const startCursor = cursor.valueOf();
        let i = 0;
        let count = 0;
        while (cursor.valueOf() < endTimestamp)
        {
            if (cursor.valueOf() >= startTimestamp)
            {
                sequence.push(cursor.valueOf());
                count++;
            }
            switch (increment)
            {
                case 'eon':
                    break;
                case '30day':
                    cursor.add(30, 'days');
                    break;
                case 'month':
                    this._incrementMomentByMonth(cursor, baseDay);
                    break;
                case 'quarter':
                    this._incrementMomentByMonth(cursor, baseDay, 3);
                    break;
                case 'halfYear':
                    this._incrementMomentByMonth(cursor, baseDay, 6);
                    break;
                case 'year':
                    cursor.add(1, 'years');
                    break;
                case 'week':
                    cursor.add(7, 'days');
                    break;
                case '2week':
                    cursor.add(14, 'days');
                    break;
                case 'arrayDays':
                    if (i < incrementArray.length)
                        cursor.add(incrementArray[i++], 'days');
                    else
                        cursor.add(incrementArray[incrementArray.length - 1], 'days');
                    break;
                default:
                    throw `Span increment '${increment}' not supported!`;
            }
            if (count >= maxCount)
                break;
        }
        if (sequence[0] !== startTimestamp)
            sequence.unshift(startTimestamp);

        if (returnIntervals)
        {
            sequence = sequence.map(x => ({ startTimestamp: x }));
            for (let i = 1; i < sequence.length; i++)
                sequence[i - 1].endTimestamp = sequence[i].startTimestamp;
            sequence[sequence.length - 1].endTimestamp = endTimestamp;
        }
        if (returnMetadata)
            return { sequence, startCursor, endCursor: cursor.valueOf() };        
        else
            return sequence;
    }
    // Return a series of timestamps from start to end incrementing using the increment key
    // Can also pass an array of day offsets like [60, 30]. If offsets run out, the last
    // last will be repeated as needed.
    // Deprecated: Use getDateSequence instead.
    getDateSpan(startTimestamp, endTimestamp, increment = 'month', maxCount = 1000000)
    {
        return this.getDateSequence(startTimestamp, endTimestamp, { increment, maxCount, returnIntervals: false, returnMetadata: false });
    }
    addToTimestamp(timestamp, count, increment = 'days')
    {
        return this._toMoment(timestamp).add(count, increment)
                                        .valueOf();
    }
    getEndOfDayTimestamp(timestamp)
    {
        return this._toMoment(timestamp).endOf('day')
                                        .valueOf();
    }
    dayCount(startTimestamp, endTimestamp)
    {
        const startM = this._toMoment(startTimestamp);
        const endM = this._toMoment(endTimestamp);

        return endM.diff(startM, 'days', true);
    }
    dayCountWhole(startTimestamp, endTimestamp)
    {
        const startM = this._toMoment(startTimestamp);
        const endM = this._toMoment(endTimestamp);

        return endM.diff(startM, 'days');
    }
    linearRatio(startTimestamp, splitTimestamp, endTimestamp, clampZeroToOne = true)
    {
        if (startTimestamp >= endTimestamp)
            throw "Linear Ratio: Denominator cannot be zero or negative.";
        
        if (splitTimestamp > endTimestamp)
            throw "Linear Ratio: Split cannot be after end.";
        
        if (splitTimestamp <= startTimestamp)
            return 0;

        let result = (splitTimestamp - startTimestamp) / (endTimestamp - startTimestamp);

        if (clampZeroToOne)
            result = Math.max(0, Math.min(1, result));

        return result;
    }
    incrementMonth(startTimestamp, numMonths = 1)
    {
        return this._toMoment(startTimestamp).add(numMonths, 'months').valueOf();
    }
    monthCount(startTimestamp, endTimestamp)
    {
        let curM = this._toMoment(startTimestamp);
        
        let startDayOfMonth;

        if (this._baseTimestamp)
            startDayOfMonth = this._toMoment(this._baseTimestamp).date();
        else
            startDayOfMonth = curM.date();

        if (startDayOfMonth < 29)
            startDayOfMonth = null;

        let count = 0;
        const yearsGap = Math.floor(endTimestamp - startTimestamp) / 31622400000; // 366 days. See how many years we can add as a single op
        if (yearsGap)
        {
            curM.add(yearsGap, 'years');
            count = yearsGap * 12;
        }
        while (curM.valueOf() < endTimestamp)
        {
            const nextM = this._incrementMomentByMonth(curM.clone(), startDayOfMonth, 1);

            if (nextM.valueOf() > endTimestamp)
                count += (endTimestamp - curM.valueOf()) / (nextM.valueOf() - curM.valueOf());
            else
                count++;
            curM = nextM;
        }
        return count;
    }
    /* Intended to duplicate the Socotra platform month count algorithm */
    socotraMonthCount(startTimestamp, endTimestamp)
    {
        const baseTimestamp = this._baseTimestamp || startTimestamp;

        const startM = this._toMoment(startTimestamp);
        const endM = this._toMoment(endTimestamp);
        const baseM = this._toMoment(baseTimestamp);

        const startDate = startM.date();
        const endDate = endM.date();

        // Simple case
        if ((baseTimestamp <= startTimestamp) && (startTimestamp < endTimestamp) && (startDate === endDate) && (startM.hour() === endM.hour()) && ((startTimestamp % 3600000) === (endTimestamp % 3600000)))
        {
            const differenceInYears = endM.year() - startM.year();
            const differenceInMonths = endM.month() - startM.month();
            return differenceInMonths + differenceInYears * 12;
        }
    
        // Normal case
        const baseDate = baseM.date();

        let monthCount = 0;
        let prevMonth = startM.clone();
        let monthCursor = this._socotraIncrementMomentByMonth(startM, baseDate);
        
        while (monthCursor.valueOf() < endTimestamp)
        {
            prevMonth = monthCursor;
            monthCursor = this._socotraIncrementMomentByMonth(monthCursor, baseDate);
            monthCount += 1;
            if (monthCursor.date() !== startDate)
            {
                const daysInMonth = monthCursor.daysInMonth();
                if (baseDate >= daysInMonth) // special case for policies that started on 29/30/31
                {
                    monthCursor.set('date', daysInMonth);
                }
                else
                {
                    monthCursor.set('date', Math.min(startDate, daysInMonth));
                }
            }
        }
        const prevMonthVal = prevMonth.valueOf();
        if (prevMonthVal === endTimestamp)
        {
            return monthCount;
        }
        else
        {
            const remainder = endTimestamp - prevMonthVal;
            const totalNextMonthMillis = monthCursor.valueOf() - prevMonthVal;
            return this._round7(monthCount + remainder / totalNextMonthMillis);
        } 
    }
    targetAmountToSocotraYearlyRate(startTimestamp, endTimestamp, targetAmount)
    {
        const months = this.socotraMonthCount(startTimestamp, endTimestamp);
        return targetAmount / months * 12;
    }
    // Computes the duration based on the units specified with the class'
    // durationUnit. baseTimestamp is optional and is only used for monthly-based
    // calcs
    getDuration(startTimestamp, endTimestamp)
    {
        if (this._baseTimestamp && (this._baseTimestamp !== startTimestamp))
            return this._getDuration(this._baseTimestamp, endTimestamp) - this.getDuration(this._baseTimestamp, startTimestamp);
        else
            return this_getDuration(startTimestamp, endTimestamp);
    }
    _getDuration(startTimestamp, endTimestamp)
    {
        switch (this._durationUnit)
        {
            case 'days360':
                return this.dayCount360(startTimestamp,
                                        endTimestamp);
            case 'months':
                return this.monthCount(startTimestamp, endTimestamp, baseTimestamp);
            case 'socotraMonths':
                return this.socotraMonthCount(startTimestamp, endTimestamp, baseTimestamp);
            case 'days':
                return this.dayCount(startTimestamp, splitTimestamp);
            case 'wholeDays':
                return this.dayCountWhole(startTimestamp, splitTimestamp);
                break;
            case 'ms':
            default:
                return endTimestamp - startTimestamp;
        }
    }
    // Computes the ratio of the pre-split segment to the overall
    // timespan based on the units specified with the class' durationUnit
    // Returns a value bound between zero and 1. If the split is before the
    // start, returns 0. If the split is after the end, returns 1.
    getDurationRatio(startTimestamp, splitTimestamp, endTimestamp)
    {
        const num = this.getDuration(startTimestamp, splitTimestamp);
        const den = this.getDuration(startTimestamp, endTimestamp);
        if (num <= 0)
            return 0;
        if (num >= den)
            return 1;
        return num / den;
    }
    // Computes the number of days in a timespan using a 30E/360 day calendar
    dayCount360(startTimestamp, endTimestamp)
    {
        return this.dayCount360WithDateInfo(this.getDateInfo(startTimestamp), this.getDateInfo(endTimestamp));
    }
    dayCount360WithDateInfo(startDi, endDi)
    {
        if (startDi.day > 30)
            startDi.day = 30;
        if (endDi.day > 30)
            endDi.day = 30;

        return (endDi.year - startDi.year) * 360 +
               (endDi.month - startDi.month) * 30 +
               (endDi.day - startDi.day);
    }
    /* mutates startM */
    _incrementMomentByMonth(startM, anchorToDayOfMonth = null, numMonths = 1)
    {
        startM.add(numMonths, 'months');
        if (anchorToDayOfMonth)
        {
            if (startM.date() < anchorToDayOfMonth)
                startM.set('date', Math.min(startM.daysInMonth(), anchorToDayOfMonth));
            else if (startM.date() > anchorToDayOfMonth)
                startM.set('date', anchorToDayOfMonth);
        }
        return startM;
    }
    /* does not mutate startM */
    _socotraIncrementMomentByMonth(startM, baseDate)
    {
        const startDate = startM.date();
        const startDaysInMonth = startM.daysInMonth();
        const nextMonth = startM.clone().add(1, 'months');
        const startDomBeforeBaseDom = startDate < baseDate;
        const nextMonthBeforeBaseDom = nextMonth.date() < baseDate;
        if (startDate === startDaysInMonth && startDomBeforeBaseDom && nextMonthBeforeBaseDom)
        {
            return nextMonth.set('date', Math.min(nextMonth.daysInMonth(), baseDate));
        }
        else
        {
            return nextMonth;
        }
    }
    _round7(amount)
    {
        return Math.round(amount * 10000000.0) / 10000000.0;
    }
    _toMoment(timestamp)
    {
        return moment(timestamp).tz(this._timeZone);
    }
}
exports.DateCalc = DateCalc;